// Author:  Alex O, Dan D
// Date: 31/03/2015
// Version: 2

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <libspe2.h>
#include <pthread.h>

#define SPU_THREADS 16

extern spe_program_handle_t spu_mailbox;

typedef struct {
   long cellno;
   int data;
   spe_context_ptr_t spe;
} thread_arg_t;

void* ppu_pthread_function(void* thread_arg)
{
   thread_arg_t* arg = (thread_arg_t*)thread_arg;
   unsigned int entry = SPE_DEFAULT_ENTRY;

   if(spe_context_run(arg->spe, &entry, 0, (void*)arg->cellno,
      (void*)(long)arg->data, NULL) < 0) {
      perror("Failed running context");
      exit(1);
   }

   pthread_exit(NULL);
}

int main(void)
{
   int i;
   spe_context_ptr_t ctxs[SPU_THREADS];
   pthread_t threads[SPU_THREADS];
   thread_arg_t arg[SPU_THREADS];

   /* Create several SPE-threads to execute 'SPU'. */
   for(i=0; i<SPU_THREADS; i++) {
      /* Create context */
      if((ctxs[i] = spe_context_create(0, NULL)) == NULL) {
         perror("Failed creating context");
         exit(1);
      }

      /* Load program into context */
      if(spe_program_load(ctxs[i], &spu_mailbox)) {
         perror("Failed loading program");
         exit(1);
      }

      /* Send some more paramters to the SPU. */
      arg[i].cellno = i;
      arg[i].spe = ctxs[i];
      arg[i].data = 55 + i;

      /* Create thread for each SPE context */
      if(pthread_create(&threads[i], NULL, &ppu_pthread_function, (void*)&arg[i])) {
         perror("Failed creating thread");
         exit(1);
      }
   }

   /* Write mailbox entry: it's definetly one message per SPU so no blocking should occur. */
   for(i=0; i<SPU_THREADS; i++) {
      spe_in_mbox_write(ctxs[i], (unsigned int*)&(arg[i].data), 1, SPE_MBOX_ALL_BLOCKING);
      printf("[PPU] data sent to SPU# = %d\n", arg[i].data);
   }

   /* Wait for SPU-thread to complete execution. */
   for(i=0; i<SPU_THREADS; i++) {
      if(pthread_join(threads[i], NULL)) {
         perror("Failed pthread_join");
         exit(1);
      }

      /* Destroy context */
      if(spe_context_destroy(ctxs[i]) != 0) {
         perror("Failed destroying context");
         exit(1);
      }
   }

   printf("\nThe program has successfully executed.\n");
   return 0;
}

